import yaml
import re
import copy
import math


class Distance(float):
    @classmethod
    def from_mm(cls, val):
        return cls(val / 1000)

    def as_mm(self):
        return float(self * 1000)

    def as_m(self):
        return float(self)

    def __add__(self, other):
        return Distance(float.__add__(self, other))

    def __sub__(self, other):
        return Distance(float.__sub__(self, other))

    def __div__(self, other):
        return Distance(float.__div__(self, other))

    def __floordiv__(self, other):
        return Distance(float.__floordiv__(self, other))

    def __mul__(self, other):
        return Distance(float.__mul__(self, other))

    def __str__(self):
        return "Distance({})".format(float(self))

    def __repr__(self):
        return str(self)

    def __neg__(self):
        return Distance(-float(self))

    def __abs__(self):
        return Distance(abs(float(self)))

distance_re = re.compile("^\s*([-\d.]+)(mm|m)\s*$")


def distance_constructor(loader, node):
    value = loader.construct_scalar(node)
    res = re.match(distance_re, value)

    val, unit = res.group(1), res.group(2)

    val = float(val)

    if unit == 'mm':
        return Distance.from_mm(val)
    elif unit == 'm':
        return Distance(val)


yaml.add_constructor(u'!distance', distance_constructor)
yaml.add_implicit_resolver(u'!distance', distance_re)


class Angle(float):
    @classmethod
    def from_degree(cls, val):
        return cls(math.radians(val))

    def as_degrees(self):
        return math.degrees(self)

    def as_radians(self):
        return float(self)

    def __str__(self):
        return "Angle({})".format(float(self))

    def __add__(self, other):
        return Angle(float.__add__(self, other))

    def __sub__(self, other):
        return Angle(float.__sub__(self, other))

    def __div__(self, other):
        return Angle(float.__div__(self, other))

    def __floordiv__(self, other):
        return Angle(float.__floordiv__(self, other))

    def __mul__(self, other):
        return Angle(float.__mul__(self, other))

    def __neg__(self):
        return Angle(-float(self))

    def __abs__(self):
        return Angle(abs(float(self)))


angle_re = re.compile(u"^\s*([-\d.]+)(\xb0*|r)\s*$")


def angle_constructor(loader, node):
    value = loader.construct_scalar(node)
    res = re.match(angle_re, value)

    val, unit = res.group(1), res.group(2)

    val = float(val)

    if unit == u'\xb0':
        return Angle.from_degree(val)
    elif unit == 'r':
        return Angle(val)
    elif unit == '':
        return Angle(val)


yaml.add_constructor(u'!angle', angle_constructor)
yaml.add_implicit_resolver(u'!angle', angle_re)


def _parse_val(val, top_cfg):

    res = re.match("^\s*%(.*)%\s*$", val)
    if res is None:
        return 0, 0, val

    expr = res.group(1)
    base_expr = expr
    vars = re.findall("[a-zA-Z][a-zA-Z0-9\._]+", expr)
    replaced_count = 0
    unreplaced_count = 0

    for var in vars:
        if var in ["max", "min", "abs", "Distance", "Angle"]:
            continue
        keys = var.split('.')
        value = top_cfg
        while len(keys) > 0:
            if isinstance(value, list):
                value = value[int(keys[0])]
            elif isinstance(value, dict):
                try:
                    value = value[keys[0]]
                except KeyError:
                    raise Exception(
                        "Using key for variable that does not exist ", keys[0],
                        " in expression: ", base_expr)
            keys = keys[1:]
        if isinstance(value, str) and re.match("^\s*%(.*)%\s*$", value):
            unreplaced_count += 1
            continue
        expr = expr.replace(var, str(value))
        replaced_count += 1

    if unreplaced_count > 0:
        expr = "%" + expr + "%"
        return replaced_count, unreplaced_count, expr

    try:
        print expr
        expr = eval(expr)
        print expr
    except:
        expr = "%" + expr + "%"

    return replaced_count, unreplaced_count, expr


# returns number of replaced values
def _parse_config_rec(cfg, top_cfg):

    keys = list()
    if isinstance(cfg, list):
        keys = range(len(cfg))

    elif isinstance(cfg, dict):
        keys = cfg.keys()

    replaced_big_count = 0
    unreplaced_big_count = 0
    for key in keys:
        if isinstance(cfg[key], (float, int, bool)):
            continue
        elif isinstance(cfg[key], str):
            replaced_count, unreplaced_count, res = _parse_val(
                cfg[key], top_cfg)
            replaced_big_count += replaced_count
            unreplaced_big_count += unreplaced_count
            cfg[key] = res
        else:
            replaced_count, unreplaced_count = _parse_config_rec(
                cfg[key], top_cfg)
            replaced_big_count += replaced_count
            unreplaced_big_count += unreplaced_count

    return replaced_big_count, unreplaced_big_count


# overlay 'a' over 'b' and return result
def overlay(a, b):

    if isinstance(a, list):
        assert isinstance(b, list)
        keys = range(len(a))
    elif isinstance(a, dict):
        assert isinstance(b, dict)
        keys = a.keys()

    res = copy.copy(b)

    for key in keys:
        if isinstance(a[key], (float, int, bool)):
            res[key] = a[key]
        elif isinstance(a[key], dict):
            try:
                res[key] = overlay(a[key], b[key])
            except KeyError:
                res[key] = a[key]
        elif isinstance(a[key], list):
            try:
                res[key] = overlay(a[key], b[key])
            except KeyError:
                res[key] = a[key]

    return res


def parse_config(filename):
    config = yaml.load(file(filename, "r"))

    while _parse_config_rec(config, config)[0] > 0:
        pass

    if _parse_config_rec(config, config)[1] > 0:
        raise Exception("Some values in config expressions were not replaced ",
                        config)

    for leg_key in config['legs'].keys():
        config['legs'][leg_key] = overlay(config['legs'][leg_key],
                                          config['default_leg'])

    return config
